//
// mpiJava version : Sung-Hoon Ko
//                   June 1998
// Northeast Parallel Architectures Center at Syracuse University
//

import java.io.*;

import mpi.*;

class PingPong {

  static final int  PINGPONGS = 48;
  static final int  SAMPLES   = 64;
  static final int  LOG2N_MAX = 10;
  static final int  N_MAX     = ( 1 << (LOG2N_MAX-3) );

  public static void main(String[] args) {
     
    double      tf[] = new double [SAMPLES + 1];
    double      t_call;
    int         i,j, j_pe, log2nbyte, nbyte, ns, n_pp, my_pe, npes;
    int         n_init, n_mess;
    //    String      Title ="Single Messages --- MPI_JAVA on Solaris(SM) ";
    String      Title ="Single Messages --- MPI_JAVA";
    PrintStream pfout = null;
    Status      status;

    MPI.Init(args);

    my_pe = MPI.COMM_WORLD.Rank();
    npes  = MPI.COMM_WORLD.Size();

/*****************************************************************************/
/*                   BYTE [nbyte*nbyte]                                      */
/*****************************************************************************/

    System.out.println("Start BYTE ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      byte A[] = new byte [nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.BYTE, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.BYTE, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.BYTE, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.BYTE, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE -- Finished");
    MPI.COMM_WORLD.Barrier();

/*****************************************************************************/
/*                   INT [nbyte*nbyte]                                       */
/*****************************************************************************/
/*
    System.out.println("Start INT ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      int A[] = new int [nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i] = i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.INT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.INT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.INT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.INT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":INT -- Finished");
    MPI.COMM_WORLD.Barrier();
*/
/*****************************************************************************/
/*                   DOUBLE [nbyte*nbyte]                                    */
/*****************************************************************************/
/*
    System.out.println("Start DOUBLE ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      double A[] = new double [nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i] = (double)i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.DOUBLE, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.DOUBLE, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.DOUBLE, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.DOUBLE, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":DOUBLE -- Finished");
    MPI.COMM_WORLD.Barrier();

*/
/*****************************************************************************/
/*                   BYTE with [1][nbyte]                                    */
/*****************************************************************************/

    System.out.println("Start BYTE with [1][nbyte*nbyte] ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      byte A[][] = new byte [1][nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, 1, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,1, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [1][nbyte*nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

/*****************************************************************************/
/*                   INT with [1][nbyte*nbyte]                               */
/*****************************************************************************/
/*
    System.out.println("Start INT with [1][nbyte*nbyte] ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      int A[][] = new int [1][nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0,1, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,1, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":INT with [1][nbyte*nbyte]-- Finished");
    MPI.COMM_WORLD.Barrier();
*/
/*****************************************************************************/
/*                   DOUBLE with [1][nbyte*nbyte]                            */
/*****************************************************************************/
/*
    System.out.println("Start DOUBLE with [1][nbyte*nbyte]...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      double A[][] = new double [1][nbyte*nbyte];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[0][i] = (double)i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, 1, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,1, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, 1, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":DOUBLE with [1][nbyte*nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();

*/
/*****************************************************************************/
/*                   BYTE with [nbyte*nbyte][1]                              */
/*****************************************************************************/
/*
    System.out.println("Start BYTE with [nbyte*nbyte][1] ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      byte A[][] = new byte [nbyte*nbyte][1];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i][0] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [nbyte*nbyte][1] -- Finished");
    MPI.COMM_WORLD.Barrier();

*/
/*****************************************************************************/
/*                   INT with [nbyte*nbyte][1]                               */
/*****************************************************************************/
/*
    System.out.println("Start INT with [nbyte*nbyte][1]...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      int A[][] = new int [nbyte*nbyte][1];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i][0] = i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0,nbyte*nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":INT with [nbyte*nbyte][1]-- Finished");
    MPI.COMM_WORLD.Barrier();
*/
/*****************************************************************************/
/*                   DOUBLE with [nbyte*nbyte][1]                            */
/*****************************************************************************/
/*
    System.out.println("Start DOUBLE with [nbyte*nbyte][1]...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      double A[][] = new double [nbyte*nbyte][1];
      
      for (i = 0; i < nbyte*nbyte; i++) {
	    A[i][0] = (double)i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte*nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte*nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte*nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":DOUBLE with [nbyte*nbyte][1]-- Finished");
    MPI.COMM_WORLD.Barrier();


*/
/*****************************************************************************/
/*                   BYTE with [nbyte][nbyte]                                */
/*****************************************************************************/

    System.out.println("Start BYTE with [nbyte][nbyte] ...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      byte A[][] = new byte [nbyte][nbyte];
      
      for (i = 0; i < nbyte; i++) {
         for(j = 0; j < nbyte; j++)
	    A[i][j] = (byte)'0'; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte*nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":BYTE with [nbyte][nbyte] -- Finished");
    MPI.COMM_WORLD.Barrier();


/*****************************************************************************/
/*                   INT with [nbyte][nbyte]                                 */
/*****************************************************************************/
/*
    System.out.println("Start INT with [nbyte][nbyte]...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      int A[][] = new int [nbyte][nbyte];
      
      for (i = 0; i < nbyte; i++) {
         for(j = 0; j < nbyte; j++)
	    A[i][j] = i; //(double) 1. / (i + 1);
      }
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0,nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":INT with [nbyte][nbyte]-- Finished");
    MPI.COMM_WORLD.Barrier();

*/
/*****************************************************************************/
/*                   DOUBLE with [nbyte][nbyte]                              */
/*****************************************************************************/
/*
    System.out.println("Start DOUBLE with [nbyte][nbyte]...");
    tf[0] = MPI.Wtime();
    for (ns = 1; ns <= SAMPLES; ++ns) {
      tf[ns] = MPI.Wtime();
    }
    
    t_call = (tf[SAMPLES] - tf[0]) / (double) SAMPLES ;

    if(my_pe==0) System.out.println("\n t_call= "+ t_call + " secs");
    
    n_init = 2 * PINGPONGS;
    n_mess = 2 * PINGPONGS;    

    if (my_pe == 0) {
      try {
	    FileOutputStream fout = new FileOutputStream("PingPong_JAVA_SM2.dat");
	    pfout = new PrintStream(fout);
      } catch(IOException ioe) {
	    System.err.println("Error in FileOutput...."+ioe);
      }
      Table.table_top(npes, n_init, n_mess, SAMPLES, t_call, Title);
    }

    for (log2nbyte = 0; log2nbyte <= LOG2N_MAX; ++log2nbyte) {
      nbyte = ( 1 << log2nbyte );
      double A[][] = new double [nbyte][nbyte];
      
      for (i = 0; i < nbyte; i++) 
         for(j = 0; j < nbyte; j++)
	    A[i][j] = (double)i; //(double) 1. / (i + 1);
         
          
      for (j_pe = 1; j_pe <= npes-1; ++j_pe) {
	    MPI.COMM_WORLD.Barrier();
	    tf[0] = MPI.Wtime();
	    for (ns = 0; ns < SAMPLES; ns++) {
	      for (n_pp = 0; n_pp < PINGPONGS; n_pp++) {
	        if (my_pe == j_pe) {
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, 0, 10);
	          status = MPI.COMM_WORLD.Recv(A, 0, nbyte, MPI.OBJECT, 0, 20);
	        }
	        if (my_pe == 0) {
	          status = MPI.COMM_WORLD.Recv(A, 0 ,nbyte, MPI.OBJECT, j_pe, 10);
	          MPI.COMM_WORLD.Send(A, 0, nbyte, MPI.OBJECT, j_pe, 20);
	        }
	      }
	    tf[ns+1] = MPI.Wtime();
	  }
	  if (my_pe == 0)
	    Table.table_body(j_pe, nbyte, n_init, n_mess, SAMPLES, t_call, tf, pfout);
      }
      
    }
    
    System.out.println("\n"+MPI.Get_processor_name()+":DOUBLE with [nbyte][nbyte]-- Finished");
    MPI.COMM_WORLD.Barrier();
*/

    MPI.Finalize();    
  }

}
